# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
from django.conf import settings
import json
import logging
from rest_framework import status


from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load, filter_networks, make_mac
)


log = logging.getLogger(__name__)


def test_network_bug_issues_34(client, site):
    """Test set queries for Networks."""

    # URIs
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    # Pre-load the attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the network objects and retreive them for testing.
    client.post(net_uri, data=load('networks.json'))
    net_resp = client.retrieve(net_uri, include_ips=True)
    net_out = net_resp.json()['data']
    networks = net_out['networks']

    # Filter networks w/ attribute hostname=foo-bar1
    expected = copy.deepcopy(net_out)
    wanted = ['192.168.0.0/24', '192.168.0.0/25']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})

    assert_success(
        client.retrieve(net_uri, attributes='hostname=foo-bar1'),
        expected
    )

    # Filter networks w/ attribute hostname=foo-bar1, including IPs
    wanted = ['192.168.0.1/32', '192.168.0.0/24', '192.168.0.0/25']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})

    assert_success(
        client.retrieve(
            net_uri, attributes='hostname=foo-bar1', include_ips=True
        ),
        expected
    )


def test_mac_address_bug_issues_111(client, site):
    """Test that a MAC coming in as an integer is properly formatted."""
    # Make sure that none of them ever match wrong.
    mac_int = 122191241314
    mac_str = '122191241314'
    mac_expected = '00:1c:73:2a:60:62'
    mac_wrong = '12:21:91:24:13:14'

    dev_uri = site.list_uri('device')
    ifc_uri = site.list_uri('interface')

    dev_resp = client.create(dev_uri, hostname='foo-bar1')
    dev = dev_resp.json()['data']['device']

    # Create the interface w/ an integer
    ifc_resp = client.create(
        ifc_uri, device=dev['id'], name='eth0', parent_id=None,
        mac_address=mac_int
    )
    ifc = ifc_resp.json()['data']['interface']
    ifc_obj_uri = site.detail_uri('interface', id=ifc['id'])

    # Test that integer matches expected
    assert make_mac(ifc['mac_address']) == mac_expected

    # Update the interface w/ a string integer
    updated = copy.deepcopy(ifc)
    updated_resp = client.put(ifc_obj_uri, data=json.dumps(updated))
    expected = updated_resp.json()['data']['interface']

    # Test that string integer matches expectd
    assert make_mac(expected['mac_address']) == mac_expected

    # And for completeness, make sure that a formatted string still comes back
    # the same.
    updated['mac_address'] = mac_expected
    updated_resp = client.put(ifc_obj_uri, data=json.dumps(updated))
    expected = updated_resp.json()['data']['interface']

    # Test that expected matches expected
    assert make_mac(expected['mac_address']) == mac_expected



"""
Validators for validating object fields.
"""

from django.conf import settings
from django.core.validators import EmailValidator
import ipaddress
from macaddress.formfields import MACAddressField as MACAddressFormField

from . import exc


def validate_mac_address(value):
    """Validate whether ``value`` is a valid MAC address."""
    if isinstance(value, basestring):
        # If the incoming value is a string, cast it to an int
        if value.isdigit():
            value = int(value)

    field = MACAddressFormField(required=False)  # allow None
    try:
        field.clean(value)
    except exc.DjangoValidationError as err:
        raise exc.ValidationError({
            'mac_address': err.message
        })

    return value


def validate_name(value):
    """Validate whether ``value`` is a valid name."""
    if not value:
        raise exc.ValidationError({
            'name': 'This is a required field.'
        })
    return value


def validate_cidr(value):
    """Validate whether ``value`` is a validr IPv4/IPv6 CIDR."""
    try:
        cidr = ipaddress.ip_network(unicode(value))
    except ValueError:
        raise exc.ValidationError({
            'cidr': '%r does not appear to be an IPv4 or IPv6 network' % value
        })
    else:
        return cidr


def validate_host_address(value):
    """Validate whether ``value`` is a host IP address."""
    cidr = validate_cidr(value)
    if cidr.prefixlen not in settings.HOST_PREFIXES:
        raise exc.ValidationError({
            'address': '%r is not a valid host address!' % value
        })
    return value


def validate_email(value):
    """Validate whether ``value`` is an email address."""
    validator = EmailValidator()
    try:
        validator(value)
    except exc.DjangoValidationError as err:
        raise exc.ValidationError({
            'email': err.message
        })
    return value



from __future__ import absolute_import, print_function

"""
Command for running any pending upgrades.
"""

from django.core.management import call_command

from nsot.util.commands import NsotCommand


class Command(NsotCommand):
    help = 'Performs any pending database migrations and upgrades'

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput',
            action='store_true',
            default=False,
            help='Tells Django to NOT prompt the user for input of any kind.',
        )

    def handle(self, **options):
        call_command(
            'syncdb',
            migrate=True,
            interactive=(not options['noinput']),
            traceback=options['traceback'],
            verbosity=options['verbosity'],
        )



from __future__ import unicode_literals

import ast
from django.contrib.auth import get_user_model
from collections import OrderedDict
import json
import logging
from rest_framework import fields, serializers
from rest_framework_bulk import BulkSerializerMixin, BulkListSerializer
import six

from . import auth
from .. import exc, models, validators
from ..util import get_field_attr


log = logging.getLogger(__name__)


###############
# Custom Fields
###############
class JSONDataField(fields.Field):
    """
    Base field used to represent attributes as JSON <-> ``field_type``.

    It is an error if ``field_type`` is not defined in a subclass.
    """
    field_type = None

    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        log.debug('JSONDictField.to_internal_value() data = %r', data)
        if self.field_type is None:
            raise NotImplementedError(
                'You must subclass JSONDataField and define field_type'
            )

        if not data:
            data = self.field_type()

        if isinstance(data, self.field_type):
            return data

        # Try it as a regular JSON object
        try:
            return json.loads(data)
        except ValueError:
            # Or try it as a Python object
            try:
                return ast.literal_eval(data)
            except (SyntaxError, ValueError) as err:
                raise exc.ValidationError(err)
        except Exception as err:
            raise exc.ValidationError(err)
        return data


class JSONDictField(JSONDataField):
    """Field used to represent attributes as JSON <-> Dict."""
    field_type = dict


class JSONListField(JSONDataField):
    """Field used to represent attributes as JSON <-> List."""
    field_type = list


class LimitedForeignKeyField(serializers.PrimaryKeyRelatedField):
    """
    Workaround for performance for when there are 100k+ related objects.

    Ref: https://github.com/tomchristie/django-rest-framework/pull/3330
    """
    @property
    def choices(self):
        return self.get_choices()

    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff:
            queryset = queryset[:cutoff]

        return OrderedDict([
            (
                six.text_type(self.to_representation(item)),
                self.display_value(item)
            )
            for item in queryset
            # for item in queryset[:self.html_cutoff]
        ])

    def iter_options(self):
        return fields.iter_options(
            # self.grouped_choices,
            self.get_choices(self.html_cutoff),
            cutoff=self.html_cutoff,
            cutoff_text=self.html_cutoff_text
        )


class MACAddressField(fields.Field):
    """Field used to validate MAC address objects as integer or string."""
    def to_representation(self, value):
        return value

    def to_internal_value(self, value):
        return validators.validate_mac_address(value)


###################
# Base Serializer #
###################
class NsotSerializer(serializers.ModelSerializer):
    """Base serializer that logs change events."""
    attributes = JSONDictField(
        required=False,
        help_text='Dictionary of attributes to set.'
    )

    def to_internal_value(self, data):
        """Inject site_pk from view's kwargs if it's not already in data."""
        kwargs = self.context['view'].kwargs

        log.debug(
            'NsotSerializer.to_internal_value() data [before] = %r', data
        )

        if 'site_id' not in data and 'site_pk' in kwargs:
            data['site_id'] = kwargs['site_pk']

        log.debug('NsotSerializer.to_internal_value() data [after] = %r', data)

        return super(NsotSerializer, self).to_internal_value(data)

    def to_representation(self, obj):
        """Always return the dict representation."""
        if isinstance(obj, OrderedDict):
            return obj
        return obj.to_dict()


######
# User
######
class UserSerializer(serializers.ModelSerializer):
    """
    UserProxy model serializer that takes optional `with_secret_key` argument
    that controls whether the secret_key for the user should be displayed.
    """
    def __init__(self, *args, **kwargs):
        # Don't pass `with_secret_key` up to the superclass
        self.with_secret_key = kwargs.pop('with_secret_key', None)
        super(UserSerializer, self).__init__(*args, **kwargs)

        # If we haven't passed `with_secret_key`, don't show the secret_key
        # field.
        if self.with_secret_key is None:
            self.fields.pop('secret_key')

    permissions = fields.ReadOnlyField(source='get_permissions')

    class Meta:
        model = get_user_model()
        fields = ('id', 'email', 'permissions', 'secret_key')


######
# Site
######
class SiteSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Site


#########
# Changes
#########
class ChangeSerializer(NsotSerializer):
    """Used for displaying Change events."""
    class Meta:
        model = models.Change


###########
# Attribute
###########
class AttributeSerializer(NsotSerializer):
    """Used for GET, DELETE on Attributes."""
    class Meta:
        model = models.Attribute
        exclude = ('attributes',)


class AttributeCreateSerializer(AttributeSerializer):
    """Used for POST on Attributes."""
    constraints = JSONDictField(required=False)
    site_id = fields.IntegerField()

    class Meta:
        model = models.Attribute
        fields = ('name', 'description', 'resource_name', 'required',
                  'display', 'multi', 'constraints', 'site_id')


class AttributeUpdateSerializer(BulkSerializerMixin,
                                AttributeCreateSerializer):
    """Used for PUT, PATCH, on Attributes."""
    class Meta:
        model = models.Attribute
        list_serializer_class = BulkListSerializer
        fields = ('id', 'description', 'required', 'display', 'multi',
                  'constraints')


#######
# Value
#######
class ValueSerializer(serializers.ModelSerializer):
    """Used for GET, DELETE on Values."""
    class Meta:
        model = models.Value
        fields = ('id', 'name', 'value', 'attribute', 'resource_name',
                  'resource_id')

    # Not sure if we want to view an attribute value w/ so much context just
    # yet.
    # def to_representation(self, obj):
    #      return obj.to_dict()


class ValueCreateSerializer(ValueSerializer):
    """Used for POST on Values."""
    class Meta:
        model = models.Value
        read_only_fields = ('id', 'name', 'resource_name')
        fields = ('id', 'name', 'value', 'attribute', 'resource_name',
                  'resource_id')


########
# Device
########
class DeviceSerializer(NsotSerializer):
    """Used for GET, DELETE on Devices."""
    class Meta:
        model = models.Device


class DeviceCreateSerializer(DeviceSerializer):
    """Used for POST on Devices."""
    site_id = fields.IntegerField()

    class Meta:
        model = models.Device
        fields = ('hostname', 'attributes', 'site_id')

    def create(self, validated_data):

        # Remove the related fields before we write the object
        attributes = validated_data.pop('attributes', {})

        # Save the base object to the database.
        obj = super(DeviceCreateSerializer, self).create(validated_data)

        # Try to populate the related fields and if there are any validation
        # problems, delete the object and re-raise the error. If not, save the
        # changes.
        try:
            obj.set_attributes(attributes)
        except exc.ValidationError:
            obj.delete()
            raise
        else:
            obj.save()

        return obj


class DeviceUpdateSerializer(BulkSerializerMixin, DeviceCreateSerializer):
    """Used for PUT, PATCH, on Devices."""
    class Meta:
        model = models.Device
        list_serializer_class = BulkListSerializer
        fields = ('id', 'hostname', 'attributes')

    def update(self, instance, validated_data):

        # Remove related fields before we write the object
        attributes = validated_data.pop('attributes', {})

        # Save the object to the database.
        obj = super(DeviceUpdateSerializer, self).update(
            instance, validated_data
        )

        # Populate the related fields and save the object, allowing any
        # validation errors to raise before saving.
        obj.set_attributes(attributes)
        obj.save()

        return obj


#########
# Network
#########
class NetworkSerializer(NsotSerializer):
    """Used for GET, DELETE on Networks."""
    class Meta:
        model = models.Network


class NetworkCreateSerializer(NetworkSerializer):
    """Used for POST on Networks."""
    cidr = fields.CharField(write_only=True)
    site_id = fields.IntegerField()

    class Meta:
        model = models.Network
        fields = ('cidr', 'attributes', 'site_id')

    def create(self, validated_data):

        # Remove the related fields before we write the object
        attributes = validated_data.pop('attributes', {})

        obj = super(NetworkCreateSerializer, self).create(validated_data)

        # Try to populate the related fields and if there are any validation
        # problems, delete the object and re-raise the error. If not, save the
        # changes.
        try:
            obj.set_attributes(attributes)
        except exc.ValidationError:
            obj.delete()
            raise
        else:
            obj.save()

        return obj


class NetworkUpdateSerializer(BulkSerializerMixin, NetworkCreateSerializer):
    """Used for PUT, PATCH on Networks."""
    class Meta:
        model = models.Network
        list_serializer_class = BulkListSerializer
        fields = ('id', 'attributes',)

    def update(self, instance, validated_data):
        log.debug('NetworkUpdateSerializer.update() validated_data = %r',
                  validated_data)

        # Remove related fields before we write the object
        attributes = validated_data.pop('attributes', {})

        # Save the object to the database.
        obj = super(NetworkUpdateSerializer, self).update(
            instance, validated_data
        )

        # Populate the related fields and save the object, allowing any
        # validation errors to raise before saving.
        obj.set_attributes(attributes)
        obj.save()

        return obj


###########
# Interface
###########
class InterfaceSerializer(NsotSerializer):
    """Used for GET, DELETE on Interfaces."""
    # parent_id = LimitedForeignKeyField(
    #     html_cutoff=100,
    #     queryset=models.Interface.objects.all(),
    parent_id = fields.IntegerField(
        required=False, allow_null=True,
        label=get_field_attr(models.Interface, 'parent', 'verbose_name'),
        help_text=get_field_attr(models.Interface, 'parent', 'help_text'),
    )

    class Meta:
        model = models.Interface


class InterfaceCreateSerializer(InterfaceSerializer):
    """Used for POST on Interfaces."""
    addresses = JSONListField(
        required=False, help_text='List of host addresses to assign.'
    )
    mac_address = MACAddressField(
        required=False, allow_null=True,
        label=get_field_attr(models.Interface, 'mac_address', 'verbose_name'),
        help_text=get_field_attr(models.Interface, 'mac_address', 'help_text'),
    )

    class Meta:
        model = models.Interface
        fields = ('device', 'name', 'description', 'type', 'mac_address',
                  'speed', 'parent_id', 'addresses', 'attributes')

    def create(self, validated_data):
        log.debug('InterfaceCreateSerializer.create() validated_data = %r',
                  validated_data)

        # Remove the related fields before we write the object
        attributes = validated_data.pop('attributes', {})
        addresses = validated_data.pop('addresses', [])

        # Forcefully set the Site for this object to that of the parent Device.
        validated_data['site'] = validated_data['device'].site

        # Save the base object to the database.
        obj = super(InterfaceCreateSerializer, self).create(validated_data)

        # Try to populate the related fields and if there are any validation
        # problems, delete the object and re-raise the error. If not, save the
        # changes.
        try:
            obj.set_attributes(attributes)
            obj.set_addresses(addresses)
        except exc.ValidationError:
            obj.delete()
            raise
        else:
            obj.save()

        return obj


class InterfaceUpdateSerializer(BulkSerializerMixin,
                                InterfaceCreateSerializer):
    "Used for PUT, PATCH on Interfaces."""
    class Meta:
        model = models.Interface
        list_serializer_class = BulkListSerializer
        fields = ('id', 'name', 'description', 'type', 'mac_address', 'speed',
                  'parent_id', 'addresses', 'attributes')

    def update(self, instance, validated_data):
        log.debug('InterfaceUpdateSerializer.update() validated_data = %r',
                  validated_data)

        # Remove related fields before we write the object
        attributes = validated_data.pop('attributes', {})
        addresses = validated_data.pop('addresses', [])

        # Save the object to the database.
        obj = super(InterfaceUpdateSerializer, self).update(
            instance, validated_data
        )

        # Populate the related fields and save the object, allowing any
        # validation errors to raise before saving.
        obj.set_attributes(attributes)
        obj.set_addresses(addresses, overwrite=True)
        obj.save()

        return obj


###########
# AuthToken
###########
class AuthTokenSerializer(serializers.Serializer):
    """
    AuthToken authentication serializer to validate username/secret_key inputs.
    """
    email = serializers.CharField()
    secret_key = serializers.CharField()

    def validate(self, attrs):
        email = attrs.get('email')
        secret_key = attrs.get('secret_key')

        if email and secret_key:
            auth_func = auth.SecretKeyAuthentication().authenticate_credentials
            user, secret_key = auth_func(email, secret_key)

            if user:
                if not user.is_active:
                    msg = 'User account is disabled.'
                    raise exc.ValidationError(msg)
                attrs['user'] = user
                return attrs
            else:
                msg = 'Unable to login with provided credentials.'
                raise exc.ValidationError(msg)
        else:
            msg = 'Must include "email" and "secret_key"'
            raise exc.ValidationError(msg)



from __future__ import absolute_import, print_function

"""
Command to start the NSoT server process.
"""

from django.conf import settings
from django.core.management import call_command
import sys

from nsot.services import http
from nsot.util.commands import NsotCommand, CommandError


class Command(NsotCommand):
    help = 'Start the NSoT server process.'

    def add_arguments(self, parser):
        parser.add_argument(
            'service',
            nargs='?',
            default='http',
            help='Starts the specified service.',
        )
        parser.add_argument(
            '--debug',
            action='store_true',
            default=False,
            help='Toggle debug output.',
        ),
        parser.add_argument(
            '--noinput',
            action='store_true',
            default=False,
            help='Tells Django to NOT prompt the user for input of any kind.',
        ),
        parser.add_argument(
            '--no-collectstatic',
            action='store_false',
            dest='collectstatic',
            default=True,
            help='Do not automatically collect static files into STATIC_ROOT.',
        ),
        parser.add_argument(
            '--no-upgrade',
            action='store_false',
            dest='upgrade',
            default=True,
            help='Do not automatically perform any database upgrades.',
        ),
        parser.add_argument(
            '-a', '--address',
            type=str,
            default='%s:%s' % (settings.NSOT_HOST, settings.NSOT_PORT),
            help='Host:port to listen on.',
        )
        parser.add_argument(
            '-k', '--worker-class',
            type=str,
            default=settings.NSOT_WORKER_CLASS,
            help='The type of gunicorn workers to use.',
        ),
        parser.add_argument(
            '-t', '--timeout',
            type=int,
            default=settings.NSOT_WORKER_TIMEOUT,
            help='Timeout before gunicorn workers are killed/restarted.',
        ),
        parser.add_argument(
            '-w', '--workers',
            type=int,
            default=settings.NSOT_NUM_WORKERS,
            help=(
                'The number of gunicorn worker processes for handling '
                'requests.'
            ),
        )

    def handle(self, **options):
        address = options.get('address')

        # Break address into host:port
        if address:
            if ':' in address:
                host, port = address.split(':', 1)
                port = int(port)
            else:
                host = address
                port = None
        else:
            host, port = None, None

        services = {
            'http': http.NsotHTTPServer,
        }

        # Ensure we perform an upgrade before starting any service.
        if options.get('upgrade'):
            print("Performing upgrade before service startup...")
            call_command(
                'upgrade', verbosity=0, noinput=options.get('noinput')
            )

        # Ensure we collect static before starting any service, but only if
        # SERVE_STATIC_FILES=True.
        if options.get('collectstatic') and settings.SERVE_STATIC_FILES:
            print("Performing collectstatic before service startup...")
            call_command('collectstatic', interactive=False, ignore=['src'])

        service_name = options.get('service')
        try:
            service_class = services[service_name]
        except KeyError:
            raise CommandError('%r is not a valid service' % service_name)

        service = service_class(
            debug=options.get('debug'),
            host=host,
            port=port,
            workers=options.get('workers'),
            worker_class=options.get('worker_class'),
            timeout=options.get('timeout'),
        )

        # Remove command line arguments to avoid optparse failures with service
        # code that calls call_command which reparses the command line, and if
        # --no-upgrade is supplied a parse error is thrown.
        sys.argv = sys.argv[:1]

        service.run()



"""
Used for caching read-only REST API responses (provided by drf-extensions).
"""

import logging
from rest_framework_extensions.key_constructor import bits, constructors
from django.core.cache import cache as djcache
from django.utils import timezone
from django.utils.encoding import force_text


log = logging.getLogger(__name__)


__all__ = ('object_key_func', 'list_key_func')


class UpdatedAtKeyBit(bits.KeyBitBase):
    """Used to store/retrieve timestamp from the cache."""
    def get_data(self, **kwargs):
        key = 'api_updated_at_timestamp'
        value = djcache.get(key, None)
        if not value:
            value = timezone.now()
            djcache.set(key, value=value)

        return force_text(value)


class ObjectKeyConstructor(constructors.DefaultKeyConstructor):
    """Cache key generator for object/detail views."""
    retrieve_sql = bits.RetrieveSqlQueryKeyBit()
    updated_at = UpdatedAtKeyBit()
    kwargs = bits.KwargsKeyBit()
    params = bits.QueryParamsKeyBit()
    unique_view_id = bits.UniqueMethodIdKeyBit()
    format = bits.FormatKeyBit()
object_key_func = ObjectKeyConstructor()


class ListKeyConstructor(constructors.DefaultKeyConstructor):
    """Cache key generator for list views."""
    list_sql = bits.ListSqlQueryKeyBit()
    pagination = bits.PaginationKeyBit()
    updated_at = UpdatedAtKeyBit()
    kwargs = bits.KwargsKeyBit()
    params = bits.QueryParamsKeyBit()
    unique_view_id = bits.UniqueMethodIdKeyBit()
    format = bits.FormatKeyBit()
list_key_func = ListKeyConstructor()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('nsot', '0004_auto_20150810_1806'),
    ]

    operations = [
        migrations.AddField(
            model_name='interface',
            name='site',
            field=models.ForeignKey(related_name='interfaces', on_delete=django.db.models.deletion.PROTECT, default=1, to='nsot.Site'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='attribute',
            name='resource_name',
            field=models.CharField(db_index=True, max_length=20, verbose_name='Resource Name', choices=[('Device', 'Device'), ('Interface', 'Interface'), ('Network', 'Network')]),
        ),
        migrations.AlterField(
            model_name='change',
            name='resource_name',
            field=models.CharField(db_index=True, max_length=20, verbose_name='Resource Type', choices=[('Network', 'Network'), ('Permission', 'Permission'), ('Attribute', 'Attribute'), ('Site', 'Site'), ('Interface', 'Interface'), ('Device', 'Device')]),
        ),
    ]
